import matplotlib.pyplot as plt

def get_input_positions(num_inputs, upper_limit):
    positions = []
    step_size = upper_limit/(num_inputs+1)
    for i in range(0, num_inputs):
        positions.append((-0.5,  step_size*(i+1)))
    return positions

def plot_neurons(layers, upper_limit, ax):
    # Code block to plot neurons spaced and centered on X
    
    x_positions_neurons = list(range(0, upper_limit, int(upper_limit/5)))
    x_positions_neurons = [i + ((upper_limit-x_positions_neurons[-1])/2) for i in x_positions_neurons]

    neuron_positions = []
    for i, layer in zip(x_positions_neurons, layers):
        neuron_positions.append([])
        # Code block to plot neurons spaced and centered on Y
        y_positions_neurons = list(range(0, layer))
        y_positions_neurons = [k + ((upper_limit-y_positions_neurons[-1])/2) for k in y_positions_neurons]


        for j in y_positions_neurons:
            neuron_positions[-1].append((i, j))
            neuron_outline = plt.Circle((i, j), 0.2, color='b',zorder=1)
            neuron = plt.Circle((i, j), 0.175, color='w', zorder=2)
            ax.add_patch(neuron_outline)
            ax.add_patch(neuron)

    return neuron_positions

def plot_conections(connections, positions_left_side, positions_rigth_side, ax):
    for i, j in connections:
        x_first_neuron = positions_left_side[i][0]
        y_first_neuron = positions_left_side[i][-1]
        x_second_neuron = positions_rigth_side[j][0]
        y_second_neuron = positions_rigth_side[j][-1]

        ax.plot((x_first_neuron, x_second_neuron), (y_first_neuron, y_second_neuron), color="b", zorder=0)

def get_antecedents_connections(antecedents):
    connections = []
    j = 0
    for i, universe in enumerate(antecedents.values()):
        for _ in universe.functions.values():
            connections.append([i, j])
            j+=1
    return connections

def get_rules_connections(active_antecedents_rules):
    connections = []
    for i, rule in enumerate(active_antecedents_rules):
        for j, num in enumerate(rule):
            if num == 1:
                connections.append([i, j])
    return connections

def get_normalisation_connections(num_rules):
    connections = []
    for i in range(0, num_rules):
        for j in range(0, num_rules):
            connections.append([i, j])
    return connections

def get_consequents_connections(num_rules):
    connections = []
    for i in range(0, num_rules):
        connections.append([i, i])
    return connections

def get_output_connections(num_rules, num_outputs):
    connections = []
    for i in range(0, num_rules):
        for j in range(0, num_outputs):
            connections.append([i, j])
    return connections

def get_neurons_per_layer(model):
    return [
        len([f for universe in model.inputs.values() for f in universe.functions.values()]),
        len(model.rules.active_antecedents_rules),
        len(model.rules.active_antecedents_rules),
        len(model.rules.active_antecedents_rules),
        len(model.outputs)
    ]

def plot_structure(model):
    num_rules = len(model.rules.active_antecedents_rules)
    num_inputs = len(model.inputs)
    num_outputs = len(model.outputs)

    fig, ax = plt.subplots(figsize=(25, 10))

    layers = get_neurons_per_layer(model)
    upper_limit = max(len(layers), max(layers))
    position = plot_neurons(layers, upper_limit, ax)

    first_layer_connections = get_antecedents_connections(model.antecedents.universes)
    second_layer_connections = get_rules_connections(model.rules.active_antecedents_rules)
    third_layer_connections = get_normalisation_connections(num_rules)
    fourth_layer_connections = get_consequents_connections(num_rules)
    fifth_layer_connections = get_output_connections(num_rules, num_outputs)

    inputs_position = get_input_positions(num_inputs, upper_limit)

    plot_conections(first_layer_connections, inputs_position, position[0], ax)
    plot_conections(second_layer_connections, position[0], position[1], ax)
    plot_conections(third_layer_connections, position[1], position[2], ax)
    plot_conections(fourth_layer_connections, position[2], position[3], ax)
    plot_conections(fifth_layer_connections, position[3], position[4], ax)

        
    ax.set_xlim(-1,  6)
    ax.set_ylim(0,  5)

    ax.axis('off')

    plt.gca().set_aspect('equal', adjustable='box')
    plt.show()


import matplotlib.pyplot as plt
import torch


def plot_universe(model, universe_name):
    universe = None
    if "Input" in universe_name:
        universe = model.inputs[universe_name]
        
    elif "Output" in universe_name:
        universe = model.outputs[universe_name]

    else:
        for univ in model.inputs.values():
            if univ.name == universe_name:
                universe = univ  
        
        for univ in model.outputs.values():
            if univ.name == universe_name:
                universe = univ 

    if not universe:
        raise ValueError(f"Please select a valid universe name")
    
    plt.title(universe.name)
    plt.margins(x=0)

    X = torch.linspace(universe.min, universe.max, 100)
        
    for function_name, function in universe.functions.items():
        Y = function(X)
        plt.plot(X.detach().numpy(), Y.detach().numpy(), label=function_name)
    
    plt.legend(loc='upper center', bbox_to_anchor=(0.5, -0.05), ncol=5)
    plt.show()


import torch

from .gauss import Gauss


class Universe(torch.nn.Module):
    """
    This class is used to define the range in which a variable
    is going to be defined in a fuzzy way, it is composed of
    several functions used to describe it. 

    Attributes
    ----------
    x : torch.Tensor
        input batched data of one variable
    name : str
        name of the universe
    merge : bool
        if True, the functions that cover simmilar area will merge
    heaviside :
        if True, the functions on the sides will become Heaviside
    universe : dict
        dict where all the functions are going to be stored
    
    Methods
    -------
    get_centers_and_intervals(n_func)
        get the centers and intervals given a max and a min
    automf(n_func)
        generate automatically gauss functions inside a universe

    Returns
    -------
    torch.tensor
        a tensor of size [n_batches, n_lines, n_functions]
    """
    def __init__(self, name: str=None, merge: bool=False, heaviside: bool=False, range: list = [None, None]) -> None:
        super().__init__()
        self._range = range
        self.min, self.max = range
        self.name = name
        self.merge = merge
        self.heaviside = heaviside
        self.functions = {}

    @property
    def range(self):
        return self._range
    
    @range.setter
    def range(self, value):
        if not isinstance(value, list) and not isinstance(value, tuple):
            raise ValueError(f"A range must be a list or tuple with 2 values, but you introduced {value}")
        if len(value) != 2:
            raise ValueError(f"A range must be a list or tuple with 2 values, but you introduced {len(value)} values")
        if not ( isinstance(value[0], int) or isinstance(value[0], float) ) or not ( isinstance(value[1], int) or isinstance(value[1], float) ):
            raise ValueError(f"Both values must be integers or floats but got {type(value[0])} and {type(value[1])}")
        if value[0]>value[1]:
            raise ValueError(f"The first value must be smaller than the second value but got {value[0]} as first value and {value[1]} as second value.")
        
        self._range = value
        self.min = self._range[0]
        self.max = self._range[1]

    def get_centers_and_intervals(self, n_func: int) -> tuple:
        interval = (self.max - self.min)/ (n_func - 1)
        return [float(self.min + interval * i) for i in range(n_func)], [float(interval) for _ in range(n_func)]

    def automf(self, n_func: int=2):
        if self.max is None or self.min is None:
            raise ValueError(f"You need to specify a range for the universe {self.name}, current range is {[self.min, self.max]}")
        
        centers, intervals = self.get_centers_and_intervals(n_func=n_func)
        self.functions = {f"Gauss_{i}": Gauss(mean=center, std=interval) for i, (center, interval) in enumerate(zip(centers, intervals))}

    def forward(self , x: torch.Tensor) -> torch.Tensor:
        if not self.functions:
            raise ValueError(f"You need to define at least one function inside the universe {self.name}. Incidentally, you can use '.automf()' to automatically generate functions.")
        
        fuzzy = None
        for _, function in self.functions.items():
            if fuzzy is None:
                fuzzy = function(x)
            else:
                fuzzy = torch.cat((fuzzy, function(x)), dim=2)
        return fuzzy


import torch

# TO BE FINISHED
class LSTSQ(torch.nn.Module):
    """
    Computes the vector x that approximately solves the equation a @ x = b

    Attributes
    ----------
    n_vars : float
        length of the "x" vector
    shoulder : float
        shoulder of the linear S function

    Returns
    -------
    torch.tensor
        a tensor of equal size to the input tensor
    """
    def __init__(self, n_vars, alpha=0.001, driver='gels') -> None:
        super().__init__()

        self.theta = torch.zeros((n_vars, 1))
        self.alpha = alpha
        self.driver = driver

        self.step = -1
        self.theta_dict = {}
        
    def forward(self, x, y=None):
        if self.training: 
            for _ in range(x.size(0)):
                self.step += 1
                theta = torch.linalg.lstsq(x, y, driver=self.driver).solution
                
                if theta.dim() > 2:
                    theta = theta.mean(dim=0)
                
                self.theta = self.theta + theta * self.alpha
                self.theta_dict[self.step] = self.theta

        #return self.theta_dict[self.step]


import torch
from .utils import init_parameter

class Gauss(torch.nn.Module):
    """
    Applies a gauss transformation to the incoming data.

    Attributes
    ----------
    mean : float
        center of the gauss function
    std : float
        width of the gauss function

    Returns
    -------
    torch.tensor
        a tensor of equal size to the input tensor
    """
    def __init__(self, mean:float = None, std:float = None) -> None:
        super().__init__()
        
        
        #if not std and not isinstance(std, float):
        #    raise ValueError(f"Expected std to be a float number but got {type(std)} instead")
        #elif std <= 0.0:
        #    raise ValueError(f"std value must be greater than 0.0 but got {std}")
    
        #if not mean and not isinstance(mean, float):
        #    raise ValueError(f"Expected mean to be a float number but got {type(std)} instead")
        
        self.mean = init_parameter(mean)
        self.std = init_parameter(std)
    
    def get_center(self) -> torch.Tensor:
        return self.mean
    
    def forward(self, x) -> torch.Tensor:
        #if not self.mean:
        #    raise RuntimeError(r"Expected mean to be a number but got None")
        #if not self.std:
        #    raise RuntimeError(r"Expected std to have a number but got None")
        
        x = x - self.mean
        x = (x)** 2
        x = -(x)/ (2 * (self.std ** 2))
        x = torch.exp(x)
        return x


import torch
from .sigmoid import Sigmoid

class Heaviside(torch.nn.Module):
    """
    This expression will be used on the corners. To indicate that a function on the
    extreme left will be 1 to the utmost left (left_equation = 1) or to the
    right (right_equation = 1). The step equation can be a sigmoid with mean on the
    edge of the transition and very little std.

    Attributes
    ----------
    left_equation : torch.tensor
        equation that will be present on the left
    rigth_equation : torch.tensor
        equation that will be present on the rigth

    Returns
    -------
    torch.tensor
        a tensor of equal size to the input tensor
    """
    def __init__(self, left_equation = None, right_equation = None) -> None:
        super().__init__()
        self.left_equation = torch.tensor([1], dtype=float) if left_equation is None else left_equation
        self.right_equation = torch.tensor([1], dtype=float) if right_equation is None else right_equation

        self.center = self.left_equation.get_center() if type(self.left_equation) != torch.Tensor else self.right_equation.get_center()
        self.step = Sigmoid(center=float(self.center), width=1e-5)

    def forward(self, x) -> torch.Tensor:
        left_equation = self.left_equation(x) if type(self.left_equation) != torch.Tensor else self.left_equation
        right_equation = self.right_equation(x) if type(self.right_equation) != torch.Tensor else self.right_equation
        return (torch.tensor(1) - self.step(x)) * left_equation +  self.step(x) * right_equation


import torch
from .utils import init_parameter

class Sigmoid(torch.nn.Module):
    """
    Applies a sigmoid transformation to the incoming data.

    Attributes
    ----------
    center : float
        center of the sigmoid function
    width : float
        width of the transition area

    Returns
    -------
    torch.tensor
        a tensor of equal size to the input tensor
    """
    def __init__(self, width:float = None, center:float = None) -> None:
        super().__init__()
        self.center = init_parameter(center)
        self.width = init_parameter(width)
    
    def get_center(self) -> torch.Tensor:
        return self.center
    
    def forward(self, x) -> torch.Tensor:
        x = x - self.center
        x = x / (- self.width)
        x = torch.exp(x)
        x = x + 1
        x = 1 / x
        return x


import torch

from pyanfis.functions import Universe

class Antecedents(torch.nn.Module):
    """
    This class is used to define the range in which a variable
    is going to be defined in a fuzzy way, it is composed of
    several functions used to describe it. 

    Attributes
    ----------
    x : torch.Tensor
        input batched data
    merge : bool
        if True, the functions that cover similar area will merge
    heaviside :
        if True, the functions on the sides will become Heaviside
    universes : dict
        dict where all the universes are going to be stored
    
    Methods
    -------
    automf(n_func)
        generate automatically gauss functions inside all universes
        inside the antecedents

    Returns
    -------
    torch.tensor
        a tensor of size [n_batches, n_lines, total_functions_of_all_universes]
    """
    def __init__(self, num_inputs: int, heaviside: bool=False) -> None:
        super(Antecedents, self).__init__()
        self.num_inputs = num_inputs
        self.heaviside = heaviside
        self.universes = {f"Input {i+1}": Universe() for i in range(num_inputs)}

    def automf(self, n_func: int=2) -> None:
        for key in self.universes.keys():
            self.universes[key].automf(n_func=n_func)

    def forward(self , X: torch.Tensor) -> torch.Tensor:
        width = len([function for key, universe in self.universes.items() for key, function in universe.functions.items()])
        fuzzy = torch.zeros(X.size(0), X.size(1), width)
        for i, (key, universe) in enumerate(self.universes.items()):
            fuzzy[:, :, i*len(universe.functions):(i+1)*len(universe.functions)] = universe(X[:,:,i:i+1])
        
        fuzzy[torch.isnan(fuzzy)] = 1
        return fuzzy


import torch

from pyanfis.consequents.types.takagi_sugeno import TakagiSugeno
from pyanfis.consequents.types.tsukamoto import Tsukamoto
from pyanfis.consequents.types.lee import Lee



CONSEQUENTS = {
    "Takagi-Sugeno": lambda num_inputs, num_outputs, parameters_update: TakagiSugeno(num_inputs, num_outputs, parameters_update),
    "Tsukamoto": lambda num_inputs, num_outputs, parameters_update: Tsukamoto(num_inputs, num_outputs, parameters_update),
    "Lee": lambda num_inputs, num_outputs, parameters_update: Lee(num_inputs, num_outputs, parameters_update),
}

class Consequents(torch.nn.Module):
    """
    This class will contain all the different types of
    consequents.

    Attributes
    ----------
    intersection : str
        intersection algorithm that is going to be used

    Methods
    -------
    generate_rules(n_membership_functions_per_universe)
        generate the rules of the universe
    relate_fuzzy_numbers(fuzzy_numbers_matrix)
        parse each input through the set of established rules

    Returns
    -------
    torch.tensor
        a tensor of size [n_batches, n_lines, n_functions]
    """
    def __init__(self, num_inputs, num_outputs, parameters_update, system_type: str="Takagi-Sugeno"):
        super().__init__()

        if not ((parameters_update == 'forward') or (parameters_update == 'backward')):
            raise ValueError("Recived {parameters_update} for parameters_update but it should be 'forward' or 'backward'.")

        self.parameters_update = parameters_update
        self.system_type = system_type
        self.consequents = CONSEQUENTS[system_type](num_inputs, num_outputs, parameters_update)

    def forward(self, f, X=None, Y=None) -> torch.Tensor:
        if self.system_type == "Takagi-Sugeno":
            return self.consequents(f,X, Y)
        else:
            return self.consequents(f)


import torch

def init_parameter(number=None):
    if number is None:
        return torch.nn.Parameter(None, requires_grad=False)
    elif isinstance(number, torch.Tensor):
        return torch.nn.Parameter(number, requires_grad=True)
    else:
        return torch.nn.Parameter(torch.tensor(number, dtype=float), requires_grad=True)


import torch

from .LSTSQ import LSTSQ
from .RLSE import RLSE

ALGORITHMS = {
    "LSTSQ": lambda n_vars: LSTSQ(n_vars),
    "RLSE":  lambda n_vars: RLSE(n_vars),
}

class TakagiSugenoAlgorithm(torch.nn.Module):
    def __init__(self, n_vars, parameters_update,  algorithm="RLSE") -> None:
        super().__init__()
        self.name = algorithm
        self.parameters_update = parameters_update

        self.algorithm_name = algorithm
        if self.name not in ALGORITHMS:
            raise ValueError(f"Invalid algorithm name: {self.name}. Supported algorithms are {list(ALGORITHMS.keys())}")
        
        self.theta = None
        self.algorithm = None

    def generate_theta(self, n_vars):
        if self.parameters_update == "backward":
            self.theta = torch.nn.Parameter(torch.zeros((n_vars, 1), requires_grad=True))
        else:
            self.algorithm = ALGORITHMS[self.algorithm_name](n_vars)

    def forward(self, x, y=None):
        if self.parameters_update != "backward":
            self.theta = self.algorithm(x.clone().detach(), y.clone().detach())
            print(self.algorithm, self.theta)


import torch

from pyanfis.functions import Sigmoid


class Lee(torch.nn.Module):
    """
    This class will compute the learnable parameters using the Lee approach.

    Attributes
    ----------
    num_inputs : float
        number of inputs that the system will recive
    num_outputs : float
        number of outputs that the system will produce
    parameters_update : float
        how the system will update the parameters

    Returns
    -------
    dict
        a dictionary that will contain the prediction related to each output
    """
    def __init__(self, num_inputs, num_outputs, parameters_update) -> None:
        super(Lee, self).__init__()
    
    def init_buffer(self, out_vars) -> dict:
        return { 
            f"output_{i}" :  Sigmoid() for i in range(out_vars)
        }
    def forward(self):
        pass


import torch

def mamdani(FN):
    '''
    INPUT: FN es Fuzzy numbers matrix,R es rules matrix
    OUTPUT: RM es Related Matrix
    '''
    RM = torch.zeros((FN.size(0), FN.size(1), FN.size(2), 1))
    for b, _ in enumerate(FN):
        for rules, _ in enumerate(FN[b]):
            for i, _ in enumerate(FN[b][rules]):
                try:
                    RM[b][rules][i] = torch.prod(FN[b, rules, i, :][FN[b, rules, i, :] > 0])
                except:
                    RM[b][rules][i] = torch.tensor([0.0])

        return RM
    
def larsen(FN):        
    '''
    INPUT: FN es Fuzzy numbers matrix,R es rules matrix
    OUTPUT: RM es Related Matrix
    '''

    mask = FN > 0
    FN_positive = FN.masked_fill(~mask, float('inf'))
    min_values, _ = FN_positive.min(dim=-1, keepdim=True)
    min_values[min_values == float('inf')] = 0

    return min_values.view(FN.size(0), FN.size(1), FN.size(2), 1)


import torch

from pyanfis.functions import Universe


class Tsukamoto(torch.nn.Module):
    """
    This class will compute the learnable parameters using the Tsukamoto approach.

    Attributes
    ----------
    num_inputs : float
        number of inputs that the system will recive
    num_outputs : float
        number of outputs that the system will produce
    parameters_update : float
        how the system will update the parameters

    Returns
    -------
    dict
        a dictionary that will contain the prediction related to each output
    """
    def __init__(self, num_inputs, num_outputs, parameters_update) -> None:
        super().__init__()
        self.universes = {f"Output {i+1}": Universe() for i in range(num_outputs)}
        # Aqui tengo que meter sigmoides para cada output
        self.parameters_update = parameters_update
        self.num_outputs = num_outputs
        self.active_rules = None

    def forward(self, f):
        #outputs = {f"Output {i+1}": torch.zeros(f.size(0), f.size(1), 1) for i in range(self.num_outputs)}
        outputs = torch.zeros(f.size(0), f.size(1), self.num_outputs)
        for j, universe in enumerate(self.universes.values()):
            X = torch.linspace(universe.min, universe.max, 200)
            functions_list = []
            for name, function in universe.functions.items():
                functions_list.append(function(X))

            function_rules = None
            for rule in self.active_rules:
                main_function = None
                for i, num in enumerate(rule):
                    if num == 1:
                        main_function = functions_list[i].unsqueeze(0) if main_function is None else torch.max(main_function, functions_list[i]).unsqueeze(0)

                function_rules = main_function if function_rules is None else torch.cat((function_rules, main_function), dim=0)
            
            for b, batch in enumerate(f):
                for i, row in enumerate(batch):
                    Y = torch.min(function_rules, row.view(-1, 1))  
                    outputs[b, i, j] = torch.sum(X * Y) / torch.sum(Y)
        return outputs


import torch
import math

from pyanfis.antecedents import Antecedents
from pyanfis.rules import Rules
from pyanfis.consequents import Consequents

from time import time

class ANFIS(torch.nn.Module):
    def __init__(self, num_inputs:int, num_outputs:int, system_type:str="Takagi-Sugeno", parameters_update:str = 'backward',):
        super().__init__()
        self.num_inputs = num_inputs
        self.num_outputs = num_outputs
        self.system_type = system_type

        self.parameters_update = parameters_update

        self.antecedents = Antecedents(num_inputs)
        self.rules = Rules()
        self.normalisation = torch.nn.functional.normalize
        self.consequents = Consequents(num_inputs=num_inputs, num_outputs=num_outputs, parameters_update=self.parameters_update, system_type=self.system_type)

        self.active_rules = None
        self.active_rules_consequents = None
        self.rules_relevancy = None
        self.erase_irrelevant_rules = None


        # The next to are pointers
        self.inputs = self.antecedents.universes # To make renaming easier
        self.outputs = self.consequents.consequents.universes # To make renaming easier

        self.firing_strength = None

    def _create_binari_rule_from_indexes(self, is_pairs, rule_index):
        rule_list = []
        for universe_name, function_name in is_pairs:
            try:
                index = rule_index.index(f"{universe_name} {function_name}")
                rule_list.append(index)
            except ValueError:
                if universe_name in [i.name for i in self.inputs.values()]:
                    raise ValueError(f"Function {function_name} not found in {universe_name}")
                else:
                    raise ValueError(f"Universe {universe_name} not found in universe list")

        rule_tensor = torch.zeros(len(rule_index))
        rule_tensor[rule_list] = 1
        return rule_tensor.unsqueeze(0)

    def _parse_rule(self, rule):
        if self.system_type == "Takagi-Sugeno" and "then" in rule:
            raise ValueError(f"Takagi-Sugeno systems only reference to the antecedent: 'If VAR1 is VALUE1 | If VAR2 is VALUE2 and VAR3 is VALUE3 | ...' the existance of a 'then' in the sencente does not make sense")

        rule = rule.split()
        antecedents_rule_index = [f"{item.name} {subkey}" for key, item in self.antecedents.universes.items() for subkey, _ in item.functions.items()]

        is_word_pairs = [(rule[i-1], rule[i+1]) for i, word in enumerate(rule) if word == 'is']

        if self.system_type != "Takagi-Sugeno":
            consequets_rule_index = [f"{item.name} {subkey}" for key, item in self.consequents.consequents.universes.items() for subkey, _ in item.functions.items()]
            
            antecedent_rules = is_word_pairs[:-1]
            consequent_rules = is_word_pairs[-1:]
            then_word_index = [i for i, word in enumerate(rule) if word == 'then']

            if rule[0] != "If" and "is" not in rule and "then" not in rule and len(then_word_index) != 1 and any(then_word_index[0] > num for num in is_word_pairs):
                raise ValueError(f"Every string containing a rule must be formated as: 'If VAR1 is VALUE1 and ... then VAR2 is VALUE2'")
            
            antecedents_rules = self._create_binari_rule_from_indexes(antecedent_rules, antecedents_rule_index)
            consequent_rules = self._create_binari_rule_from_indexes(consequent_rules, consequets_rule_index)

            return antecedents_rules, consequent_rules
        else:
            antecedents_rules = self._create_binari_rule_from_indexes(is_word_pairs, antecedents_rule_index)
            return antecedents_rules

    def create_rules_base(self, rules):
        if not isinstance(rules, torch.Tensor) and not isinstance(rules, list):
            raise ValueError(f"The introduced rules must be either a torch.Tensor or a list")
        
        if self.system_type == "Takagi-Sugeno":
            for rule in rules:
                antecedent_part = self._parse_rule(rule)
                if self.active_rules is None:
                    self.active_rules = antecedent_part
                else:
                    self.active_rules = torch.cat((self.active_rules, antecedent_part), dim=0)

            for algorithm in self.outputs.values():
                algorithm.generate_theta(self.active_rules.size(0) * (self.num_inputs + 1))
                
        elif self.system_type == "Tsukamoto":
            for rule in rules:
                antecedent_part, consequent_part = self._parse_rule(rule)
                if self.active_rules is None and self.active_rules_consequents is None:
                    self.active_rules = antecedent_part
                    self.active_rules_consequents = consequent_part
                elif self.active_rules is not None and self.active_rules_consequents is not None:
                    self.active_rules = torch.cat((self.active_rules, antecedent_part), dim=0)
                    self.active_rules_consequents = torch.cat((self.active_rules_consequents, consequent_part), dim=0)
                else:
                    raise ValueError(f"Got {len(self.active_rules)} antecedent statements and {len(self.active_rules_consequents)} consequent statements")


        elif self.system_type == "Lee":
            pass

        self.rules.active_antecedents_rules = self.active_rules

    def parameters(self):
        
        parameters = []

        # Antecedents parameters
        for universe in self.antecedents.universes.values():
            for function in universe.functions.values():
                for param in function.parameters():
                    parameters.append(param)

        # Consequent parameters
        if self.parameters_update == "backward":
            for universe in self.consequents.consequents.universes.values():
                if self.system_type == "Takagi-Sugeno":
                    parameters.append(universe.theta)
                else:
                    for function in universe.functions.values():
                        for param in function.parameters():
                            parameters.append(param)
        #print([round(param.data.item(), 2) for param in parameters])
        return parameters

    def smart_concat(self, tensor_list):
        dimensions = tensor_list[0].dim()
        shape = tensor_list[0].shape

        tensor_list = torch.stack(tensor_list, dim=-1)
        if dimensions == 0:
            return tensor_list.view(1,1,-1)
        elif dimensions == 1:
            return tensor_list.view(1,shape[0],-1)
        elif dimensions == 2:
            return tensor_list.view(shape[0], shape[1],-1)
        elif dimensions == 3:
            return tensor_list.view(shape[0], shape[1] * shape[2] , -1) 
        
    def _prepare_kwargs_matrices(self, **kwargs):
        antecedents_tensor = []
        for universe in self.antecedents.universes.values():
            if universe.name not in list(kwargs.keys()):
                raise ValueError(f"Universe name {universe.name} not present in input variables {list(kwargs.keys())}")
            antecedents_tensor.append(kwargs[universe.name])
            del kwargs[universe.name]
        
        if not kwargs and self.system_type == "Takagi-Sugeno" and self.training is True and self.parameters_update != "backward":
            raise ValueError(f"If you use a {self.system_type} and do not update the system using backpropagation you need to feed the output values to train the system.")
        elif not kwargs:
            return self.smart_concat(antecedents_tensor), None
         
        consequents_tensor = []
        for universe in self.consequents.consequents.universes.values():
            if universe.name not in list(kwargs.keys()):
                raise ValueError(f"Universe name {universe.name} not present in input variables {list(kwargs.keys())}")

            consequents_tensor.append(kwargs[universe.name])
            del kwargs[universe.name]

        return self.smart_concat(antecedents_tensor), self.smart_concat(consequents_tensor)

    def get_fired_rules(self, **kwargs):
        self.training = False
        X, _ = self._prepare_kwargs_matrices(**kwargs)
        if X.size(1) != 1:
            raise ValueError(f"Only one row can be evaluated at a time")

        f = self.antecedents(X)

        self.rules.active_rules = self.active_rules
        f, _ = self.rules(f) # col_indexes = rule place on each col
        
        f = self.normalisation(f, dim=2, p=1)

        return {str(key.to(torch.int16).tolist()): float(strength) for key, strength in zip(self.active_rules,f[0, 0, :])}
    
    def state_dict(self):
        params = {}

        # Main
        params["main"] = {}
        params["main"]["num_inputs"] = self.num_inputs
        params["main"]["num_outputs"] = self.num_outputs
        params["main"]["system_type"] = self.system_type
        params["main"]["parameters_update"] = self.parameters_update
        params["main"]["active_rules"] = self.active_rules
        params["main"]["rules_relevancy"] = self.rules_relevancy
        params["main"]["active_rules_consequents"] = self.active_rules_consequents
        params["main"]["erase_irrelevant_rules"] = self.erase_irrelevant_rules
        params["main"]["firing_strength"] = self.firing_strength

        # Antecedents
        params["Antecedents"] = {}
        params["Antecedents"]["universes"] = {}


        # Each universe in the antecedents
        for universe_name, universe in self.inputs.items():
            params["Antecedents"]["universes"][universe_name] = {}
            params["Antecedents"]["universes"][universe_name]["name"] = universe.name
            params["Antecedents"]["universes"][universe_name]["min"] = universe.min
            params["Antecedents"]["universes"][universe_name]["max"] = universe.max
            params["Antecedents"]["universes"][universe_name]["merge"] = universe.merge
            params["Antecedents"]["universes"][universe_name]["heaviside"] = universe.heaviside
            params["Antecedents"]["universes"][universe_name]["functions"] = {}

            for function_name, function in universe.functions.items():
                params["Antecedents"]["universes"][universe_name]["functions"][function_name] = {}
                params["Antecedents"]["universes"][universe_name]["functions"][function_name]["type"] = str(function)[:-2]
                params["Antecedents"]["universes"][universe_name]["functions"][function_name]["parameters"] = {}
                for name, value in vars(function)['_parameters'].items():
                    params["Antecedents"]["universes"][universe_name]["functions"][function_name]["parameters"][name] = value

        # Consequents
        params["Consequents"] = {}
        params["Consequents"]["universes"] = {}

        for universe_name, universe in self.outputs.items():
            params["Consequents"]["universes"][universe_name] = {}
            params["Consequents"]["universes"][universe_name]["name"] = universe.name
            params["Consequents"]["universes"][universe_name]["min"] = universe.min
            params["Consequents"]["universes"][universe_name]["max"] = universe.max
            params["Consequents"]["universes"][universe_name]["merge"] = universe.merge
            params["Consequents"]["universes"][universe_name]["heaviside"] = universe.heaviside

            # Takagi Sugeno
            if self.system_type == "Takagi-Sugeno":
                params["Consequents"]["universes"][universe_name][universe.algorithm_name] = {}
                params["Consequents"]["universes"][universe_name][universe.algorithm_name]["name"] = universe.name
                params["Consequents"]["universes"][universe_name][universe.algorithm_name]["theta"] = universe.theta
                params["Consequents"]["universes"][universe_name][universe.algorithm_name]["algorithm"] = universe.algorithm
                

            # Tsukamoto and Lee
            else:
                params["Consequents"]["universes"][universe_name]["functions"] = {}
                for function_name, function in universe.functions.items():
                    params["Consequents"]["universes"][universe_name]["functions"][function_name] = {}
                    params["Consequents"]["universes"][universe_name]["functions"][function_name]["type"] = str(function)[:-2]
                    params["Consequents"]["universes"][universe_name]["functions"][function_name]["parameters"] = {}
                    for name, value in vars(function)['_parameters'].items():
                        params["Consequents"]["universes"][universe_name]["functions"][function_name]["parameters"][name] = value
        return params
    
    def load_state_dict(self, state_dict):
        # Load Main
        if self.num_inputs != state_dict["main"]["num_inputs"]:
            raise ImportError(f"Atempting to import a system with {state_dict['main']['num_inputs']} inputs into a created system of {self.num_inputs} inputs.")
        
        if self.num_outputs != state_dict["main"]["num_outputs"]:
            raise ImportError(f"Atempting to import a system with {state_dict['main']['num_outputs']} outputs into a created system of {self.num_outputs} outputs.")
        
        self.system_type = state_dict["main"]["system_type"]
        self.parameters_update = state_dict["main"]["parameters_update"]

        self.antecedents = Antecedents(self.num_inputs)
        self.rules = Rules()
        self.normalisation = torch.nn.functional.normalize
        self.consequents = Consequents(num_inputs=self.num_inputs, num_outputs=self.num_outputs, parameters_update=self.parameters_update, system_type=self.system_type)

        self.active_rules = state_dict["main"]["active_rules"]
        self.active_rules_consequents = state_dict["main"]["active_rules_consequents"]
        self.rules_relevancy = state_dict["main"]["rules_relevancy"]
        self.erase_irrelevant_rules = state_dict["main"]["erase_irrelevant_rules"]

        self.firing_strength = state_dict["main"]["firing_strength"]

        # Load Antecedents
        for universe_name, universe in self.antecedents.universes.items():
            universe.name = state_dict["Antecedents"]["universes"][universe_name]["name"]
            universe.min = state_dict["Antecedents"]["universes"][universe_name]["min"]
            universe.max = state_dict["Antecedents"]["universes"][universe_name]["max"]
            universe.merge = state_dict["Antecedents"]["universes"][universe_name]["merge"]
            universe.heaviside = state_dict["Antecedents"]["universes"][universe_name]["heaviside"]

            # Loading functions
            for function_name, function_params in state_dict["Antecedents"]["universes"][universe_name]["functions"].items():
                function_type = function_params["type"]
                function_params = function_params["parameters"]
                try:
                    module = __import__("pyanfis.functions", fromlist=[function_type])
                    universe.functions[function_name] =  getattr(module, function_type)()
                except ImportError:
                    raise ImportError(f"Error: Class {function_type} not found in the 'functions' folder.")

                for name, value in function_params.items():
                    universe.functions[function_name]._parameters[name] = value

        # Load Consequents
        for universe_name, universe in self.consequents.consequents.universes.items():
            universe.name = state_dict["Consequents"]["universes"][universe_name]["name"]
            universe.min = state_dict["Consequents"]["universes"][universe_name]["min"]
            universe.max = state_dict["Consequents"]["universes"][universe_name]["max"]
            universe.merge = state_dict["Consequents"]["universes"][universe_name]["merge"]
            universe.heaviside = state_dict["Consequents"]["universes"][universe_name]["heaviside"]

            # Takagi Sugeno
            if self.system_type == "Takagi-Sugeno":
                universe.name = state_dict["Consequents"]["universes"][universe_name][universe.algorithm_name]["name"]
                universe.theta = state_dict["Consequents"]["universes"][universe_name][universe.algorithm_name]["theta"]
                universe.algorithm = state_dict["Consequents"]["universes"][universe_name][universe.algorithm_name]["algorithm"]

            # Tsukamoto and Lee
            else:
                # Loading functions
                for function_name, function_params in state_dict["Consequents"]["universes"][universe_name]["functions"].items():
                    function_type = function_params["type"]
                    function_params = function_params["parameters"]

                    try:
                        module = __import__("pyanfis.functions", fromlist=[function_type])
                        universe.functions[function_name] =  getattr(module, function_type)()
                    except ImportError:
                        raise ImportError(f"Error: Class {function_type} not found in the 'functions' folder.")

                    for name, value in function_params.items():
                        universe.functions[function_name]._parameters[name] = value
        
        # The next to are pointers
        self.inputs = self.antecedents.universes # To make renaming easier
        self.outputs = self.consequents.consequents.universes # To make renaming easier

    def _irrelevant_rules_check(self, f):
        relevancy = torch.mean(torch.mean(f, dim=0), dim=0)
        if self.rules_relevancy is None:
            self.rules_relevancy = relevancy
        else:
            self.rules_relevancy += relevancy

        self.rules_relevancy = torch.nn.functional.normalize(self.rules_relevancy, dim=0, p=1)

    def _sanity_check(self, matrix):
        '''
        Will check if a matrice is batched
        '''
        if matrix is None:
            return None
        
        if matrix.dim() == 2:
            return matrix[None, :, :]
        elif matrix.dim() == 3:
            return matrix
        elif matrix.dim() > 3 or matrix.dim() < 2:
            raise ValueError(f"Expected a matrix with 2 or 3 dimensions but got one with {matrix.dim()}")
        

    def forward(self, X, Y):

        f = self.antecedents(X)

        self.rules.active_antecedents_rules = self.active_rules
        f = self.rules(f)

        f = self.normalisation(f, dim=2, p=1)

        self.consequents.consequents.active_rules = self.active_rules_consequents
        output = self.consequents(f, X, Y)

        return output
    
    def _prepare_args_matrices(self, args):
        if len(args) > 2:
            raise ValueError("Please provide as input either a matrix with the input arguments or two matrices, one with input arguments and one with objective arguments")

        if self.parameters_update == "backward" and len(args) >= 2:
            raise ValueError(f"The selected propagation is {self.parameters_update} but you provided more than one tensor as input. Please, join all the input tensors before feeding them into the system.") 

        if self.parameters_update == "backward" or len(args) == 1:
            return args[0], None
        
        else:
            return args[0], args[1]

    def __call__(self, *args, **kwargs):
        if args and kwargs:
            raise ValueError("All the arguments must be either arguments or keyword arguments, but you cannot mix between both")

        if args and not isinstance(args[0], torch.Tensor):
            raise ValueError(f"Expected torch.Tensor as input but recived {type(args)}")
        
        if kwargs:
            X, Y = self._prepare_kwargs_matrices(**kwargs)
        
        else:
            X, Y = self._prepare_args_matrices(args)
            
        X, Y = self._sanity_check(X), self._sanity_check(Y)

        return self.forward(X, Y)


import torch

from .intersection_algorithms import larsen, mamdani

INTERSECTIONS = {
    'larsen': larsen,
    'mamdani': mamdani,
}

class Rules(torch.nn.Module):
    """
    This class will contain all the rules of the system,
    it will dictate how each one of the antecedent functions
    relate with each other.

    Attributes
    ----------
    intersection : str
        intersection algorithm that is going to be used

    Methods
    -------
    generate_rules(n_membership_functions_per_universe)
        generate the rules of the universe
    relate_fuzzy_numbers(fuzzy_numbers_matrix)
        parse each input through the set of established rules

    Returns
    -------
    torch.tensor
        a tensor of size [n_batches, n_lines, n_functions]

    Examples
    --------
    """
    def __init__(self, intersection:str = 'larsen'):
        super(). __init__()
        self.active_antecedents_rules = None
        self.active_consequents_rules = None
        self.intersection = INTERSECTIONS[intersection]
    
    def relate_fuzzy_numbers(self, fuzzy_numbers_matrix):
        '''
        INPUT: FN es Funny numbers matrix y R es rules matrix
        OUTPUT: FA Fuzzy And matrix
        '''
        fuzzy_numbers_matrix_expanded = fuzzy_numbers_matrix.unsqueeze(2)
        active_antecedents_rules_expanded = self.active_antecedents_rules.unsqueeze(0).unsqueeze(0)

        # Perform element-wise multiplication with broadcasting
        return fuzzy_numbers_matrix_expanded * active_antecedents_rules_expanded
        

    def binarice(self, binary_list: torch.Tensor) -> str:
        return str(int(''.join(str(int(i)) for i in binary_list), 2))
    
    def forward(self, x):
        x = self.intersection(self.relate_fuzzy_numbers(x)) # This is a 4D tensor
        return x[:, :, :, 0]


import torch

class RLSE(torch.nn.Module):
    """
    Computes the vector x that approximately solves the equation a @ x = b
    using a recursive approach

    Attributes
    ----------
    n_vars : float
        length of the "x" vector
    initial_gamma : float
        big number to initialise the "S" matrix

    Returns
    -------
    torch.tensor
        a tensor of equal size to the input tensor
    """
    def __init__(self, n_vars, initial_gamma=1000.):
        super().__init__()
        self.S = torch.eye(n_vars, dtype=float) * initial_gamma
        self.theta = torch.zeros((n_vars, 1), dtype=float)
        self.gamma = 1000.0

    def forward(self, A, B,):
        batch, row, _ = A.size()

        for ba in range(batch):
            for i in range(row):
                a = A[ba, i, :].view(1, -1)  # Reshape a to match the dimensions for matrix operations
                b = B[ba, i].unsqueeze(0)
                
                self.S = self.S - (torch.matmul(torch.matmul(torch.matmul(self.S, a.T), a), self.S)) / (1 + torch.matmul(torch.matmul(a, self.S), a.T))
                self.theta =  self.theta + torch.matmul(self.S, torch.matmul(a.T, (b - torch.matmul(a, self.theta))))

        return self.theta


from .universe import Universe
from .sigmoid import Sigmoid
from .gauss import Gauss
from .bell import Bell
from .linear_z import LinearZ
from .linear_s import LinearS
from .heaviside import Heaviside

__all__ = ["Universe", "Sigmoid", "Gauss", "Bell", "LinearZ", "LinearS", "Heaviside"]


from .types.takagi_sugeno import TakagiSugeno
from .types.tsukamoto import Tsukamoto
from .types.lee import Lee

from .consequents import Consequents

__all__ = ["Consequents", "TakagiSugeno", "Tsukamoto", "Lee"]


import torch

from pyanfis.algorithms import TakagiSugenoAlgorithm 


class TakagiSugeno(torch.nn.Module):
    """
    This class will compute the learnable parameters using the Takagi-Sugeno approach.

    Attributes
    ----------
    num_inputs : float
        number of inputs that the system will recive
    num_outputs : float
        number of outputs that the system will produce
    parameters_update : float
        how the system will update the parameters

    Returns
    -------
    dict
        a dictionary that will contain the prediction related to each output
    """
    def __init__(self, num_inputs, num_outputs, parameters_update) -> None:
        super().__init__()
        self.universes = {f"Output {i+1}" : TakagiSugenoAlgorithm(num_inputs, parameters_update) for i in range(num_outputs)}
        self.parameters_update = parameters_update
        self.num_outputs = num_outputs
        
    def forward(self,f, X=None, Y=None):

        ones = torch.ones(X.shape[:-1] + (1,), dtype=X.dtype)
        X = torch.cat([X, ones], dim=-1)

        x_b, x_i, _ = X.size()

        #output = {f"Output {i+1}": torch.zeros((x_b, x_i , 1)) for i in range(self.num_outputs)}
        outputs = torch.zeros(f.size(0), f.size(1), self.num_outputs)
        X = torch.einsum('bri, brj -> brij', f, X).view(x_b, x_i, -1)

        for i, (algorithm) in enumerate(self.universes.values()):
            if Y is not None:
                algorithm(X, Y[:, :, i:i+1])

            outputs[:, :, i:i+1] = outputs[:, :, i:i+1] + torch.einsum('bij, jk -> bik', X.float(), algorithm.theta.float())
   
        return outputs


import torch
from .utils import init_parameter

class LinearS(torch.nn.Module):
    """
    Applies a linear S transformation to the incoming data.

    Attributes
    ----------
    foot : float
        foot of the linear S function
    shoulder : float
        shoulder of the linear S function

    Returns
    -------
    torch.tensor
        a tensor of equal size to the input tensor
    """
    def __init__(self, foot:float = None, shoulder:float = None) -> None:
        super().__init__()
        self.shoulder = init_parameter(shoulder)
        self.foot = init_parameter(foot)
    
    def get_center(self) -> torch.Tensor:
        return self.foot - self.shoulder
    
    def forward(self, x) -> torch.Tensor:
        x = x - self.foot
        x = x / (self.shoulder - self.foot)
        x = torch.maximum(x, torch.tensor(0))
        x = torch.minimum(x, torch.tensor(1))
        return x
    
    def __setitem__(self, key, value):
        if key == "shoulder":
            self.shoulder = init_parameter(value)
        elif key == "foot":
            self.foot = init_parameter(value)
        else:
            raise KeyError(f"Invalid key: {key}")


import torch
from itertools import combinations

def binarize_input(x: torch.Tensor, min_support: int = None) -> torch.Tensor:
    if min_support == None:
        min_support = torch.mean(x)

    t = x.clone().detach()
    t[t < min_support ] = 0
    t[t != 0] = 1
    return torch.reshape(t, (t.size(0) * t.size(1), t.size(2)))

def create_rules_matrix(universes_combinations: list, fuzzyfied_input: torch.Tensor) -> torch.Tensor:

    rules_tensor = torch.zeros(len(universes_combinations), fuzzyfied_input.size(1))
    for i, line in enumerate (rules_tensor):
        for j, _ in enumerate (line):
            if j in universes_combinations[i]:
                rules_tensor[i, j] = 1

    return rules_tensor

def apriori(x: torch.Tensor, min_support = 0.5, min_support_fuzzyfication = None):
    x = x.clone().detach()
    fuzzyfied_input = binarize_input(x, min_support_fuzzyfication)

    final_conv = []
    while len(final_conv) == 0:
        initial_conv = list(range(fuzzyfied_input.size(1)))
        assert min_support > 0
        for k in range (1, fuzzyfied_input.size(1) + 1):
            for i, combination in enumerate(combinations(initial_conv , k)):
                if torch.min(torch.sum(fuzzyfied_input[:, combination], axis = 0) / fuzzyfied_input.size(0)) < min_support :
                    initial_conv = initial_conv[:i] + initial_conv[i+1:]
                else:
                    final_conv.append(combination)

        min_support -= 0.1

    return create_rules_matrix(final_conv, fuzzyfied_input)